%
% AdVirgo IMC simulation
% This is a simple version which does not include back scattering
% 
% example usage follows
%
%{
% define IMC
imc = AdVirgoIMCIFO;
% lock the cavity
imc.lock;
% print various information
imc.printAll;
% field on HR size on MC1, coming from MC2. This ie before shield
figure(1)
[E1,x] = imc.getField('MC1-HRy-i');
pcolor( x, x, log(abs(E1)) ); shading flat; axis equal; colorbar;
% field on HR size on MC3, going out to MC2. This is after shield
figure(2)
E3 = imc.getField('MC3-HRx-o');
pcolor( x, x, log(abs(E3)) ); shading flat; axis equal; colorbar;

% LG mode fraction of the field from MC1 to MC3, LG00, LG10, ..., LG50
format longe
imc.LGFrac('MC1-HRx-o', 0:5, 0)

% add thermal bump on MC1 and MC3, each absorbing 0.2W.
imc.updateMirMaps('MC1', 'PcoatMC1', 0.2);
imc.updateMirMaps('MC3', 'PcoatMC3', 0.2);
% second argument 0 means 'suppress progress inforamtion printing'
imc.lock( 1e-6, 0 );
% LG mode fraction with thermal deformation
imc.LGFrac('MC1-HRx-o', 0:5, 0)
%}
% 

%% main code
classdef AdVirgoIMCIFO < FFTIFO
    
    methods
        function obj = AdVirgoIMCIFO( varargin )
            obj@FFTIFO( varargin{:} );
        end
        
        function defineIFO( obj, varargin )
%% parameter definitions
            AdVirgoIMCIFOParams
            
%% setup IMC optics
            % fist add optics
            obj.addLaser( 'src' );
            obj.addMirror( 'MC1', 'invRoC', 0, 'Aperture', Aperture13, ...
                'Thick', thick13, 'T', T_MC1, 'Lhr', Lhr_MC1, 'Lar', Lar_MC13, 'incAngle', theta1 );
            obj.addMirror( 'MC3', 'invRoC', 0, 'Aperture', Aperture13, ...
                'Thick', thick13, 'T', T_MC3, 'Lhr', Lhr_MC3, 'Lar', Lar_MC13, 'incAngle', theta1 );
            obj.addMirrorGeneric( MC2Mirror, 'MC2', 'invRoC', 1/RoCMC2, 'Aperture', Aperture2, 'resolution', res2, ...
                'Thick', thick2, 'T', T_MC2, 'Lhr', L_MC2, 'incAngle', theta2, 'optX0', x0, 'optY0', y0 ); % , 'virtualAperture', Aperture2+0.04 );
            obj.addDetector( 'refl' );
            obj.addDetector( 'trans' );
            
            % connect optics
            % MC1-HRx] => [MC3-HRy => MC3-HR-x] => [MC2-HRy => MC2-HRx] => [MC1-HRy
            obj.addProp( '', 'src', 'MC1-ARx' );
            obj.addProp( 'Prop13', 'MC1-HRx', 'MC3-HRy', L13 );
            obj.addProp( '', 'MC3-HRx', 'MC2-HRy', L12 );
            obj.addProp( '', 'MC2-HRx', 'MC1-HRy', L12 );
            obj.addProp( '', 'MC1-ARy', 'refl' );
            obj.addProp( '', 'MC3-ARy', 'trans' );
            
            % if shield is to be placed
            if useShield > 0
                obj.setModMap( 'MC1', @obj.DIHEDRONShieldMC1 );
                obj.setModMap( 'MC3', @obj.DIHEDRONShieldMC3 );
            end
            
%% define the details about the MC HR surface
            % setHRfiles( 'mirrorName', 'dataFile', 'formula',
            %             aperture of power fitting (0 for not specifying),
            %             RoC of the power fitting (set to 0 for no fitting),
            %             orientation of the optic (0:no orientation, 1:90 degree, etc),
            %             'variableName1 in formula', initial value of variable1, 'vN2', v2, ... );
            %    Datafile format supported as : MetroPro data, Wyko OPD,
            %       LMA COL and SIS_Nxxx_Wyyy
            %    Mirror curvatures are measuted in a certain radius. 
            %    Specify that radius and the RoC calculated in that radius.
            %       The power term is adjusted to make the RoC to match the
            %       measurement. 
            %    'formula' can be any matlab expression. Following examples
            %       calls calcTermal to implement the Hello-Vinet formula.
            %    By registering variables, like PcoatMC1, the absorption can
            %       be changed by specifying different values for the variable.
            
            obj.setHRfiles('MC1', MC1map, ...
                'map*mapWgt + calcThermal(0, r, 0, PcoatMC1, wMC1, 0.02, 0.02 )', ...
                0, 1e20, 0, 'PcoatMC1', PcoatMC1, 'wMC1', 0.0051, 'mapWgt', mapWgt );
            obj.setHRfiles('MC2', MC2map, ...
                'map*mapWgt + r.^2/2*invRoC2 + calcThermal(0, r, 0, PcoatMC2, wMC2, 0.02, 0.07 ) + x*tiltX + y*tiltY', ...
                0.130/2, RoCMC2, 0, 'PcoatMC2', PcoatMC2, 'wMC2', 0.0108, 'tiltX', tiltX, 'tiltY', tiltY, 'invRoC2', 0, 'mapWgt', mapWgt );
            obj.setHRfiles('MC3', MC3map, ...
                'map*mapWgt + calcThermal(0, r, 0, PcoatMC3, wMC3, 0.02, 0.02 )', ...
                0, 1e20, 0, 'PcoatMC3', PcoatMC3, 'wMC3', 0.0051, 'mapWgt', mapWgt );
        end
        
    end
    
    methods( Static )
%% implement the sheid in front of DIHEDRON
        % for each mirror, after transmission and reflection maps are
        % calculated, functions registered by setModMap are called for each
        % mirror. reflIn and transIn are the calcuation using the mirror
        % data. shieldMask below apply the clipping by the hole on the
        % shield.
        % refl  {1}:HRx->HRy, {2}:HRy->HRx, {3}:ARx->ARy, {4}:ARy->ARx
        
        function [reflOut, transOut] = DIHEDRONShieldMC1( reflIn, transIn, x, y, ~ )
            [reflOut, transOut] = AdVirgoIMCIFO.shieldMask( 1, reflIn, transIn, x, y );
        end
        
        function [reflOut, transOut] = DIHEDRONShieldMC3( reflIn, transIn, x, y, ~ )
            [reflOut, transOut] = AdVirgoIMCIFO.shieldMask( 3, reflIn, transIn, x, y );
        end
        
        function [reflOut, transOut] = shieldMask( whichMC, reflIn, transIn, x, y )
            global DHS_offset DHS_dx DHS_dy
            
            shieldRadius = 0.02;
            transOut = transIn;
            reflOut = cell(size(reflIn));
            reflOut(3:4) = reflIn(3:4);
            
            if whichMC == 1
                clearAperture = ( (x-DHS_dx-DHS_offset).^2 + (y-DHS_dy).^2 ) < shieldRadius^2;
                reflOut{1} = fliplr(clearAperture) .* reflIn{1};
                reflOut{2} = reflIn{2} .* clearAperture;
            else
                clearAperture = ( (x-DHS_dx+DHS_offset).^2 + (y-DHS_dy).^2 ) < shieldRadius^2;
                reflOut{1} = reflIn{1} .* fliplr(clearAperture);
                reflOut{2} = clearAperture .* reflIn{2};
            end
        end
    end
end
