%
% fieldManager
%
%   HG coefficient at fldID port, for a standard field Fld
%     Cmn = FHGCoef( obj, m, n, fldID, Fld )
%
%   Convert internal field to standard format
%     Estd = conv2std( obj, fieldID, Eint )
%
%   set the field at port fldID in the intenal format
%     setHGField( obj, fldID, m, n, Power = 1, invQ = internal value )
%
%   apply propagation by requesting the output window size with variable propagation
%     [outField, xl, yl] = getField( fieldID, Wout, Lprop )
%
% September 2nd, 2015
%   lam0 and kLam0 are stored in fieldManager object, moved from ConstantTK
%   This is modified by addLaser
%
% October 27, 2014
%   conv2std is changed to return zero matrix when the field is not ready
%
% October 26, 2014
%   handle of FFTIFO itself is saved in all manager objects, 
%   instead of handle of multiple manager handles
%
% August 29, 2014
%   function [Fld, xl, yl] = getField( fldID, Wout, Lprop ) <== used to be onDemand
%   getField has a new argument, Lprop.
%
% July 31, 2014
%   1) input beam can be LGpm or HGmn
%      setInputType( type, ind1, ind2 ) is the interface to define these
%   2) [E, xList] = fieldAtFID( fieldID, 'LG or HG', p/m, m/n ) added to get field of specified type
%      fieldAtFID(fieldID, no field specification) returns using the type specified at addLaser
%
%
% July 28, 2014
%   findNamedField ( obj, name, check ) when check < 0, calcIfNecessary is not called

classdef fieldManager < handle
    
    properties
        ifo

        fieldList
        
        % CR field specification. These two need to be set properly.
        % aetherically better to use set/get, but, for now, these are set
        % in addLaser by hand
        lamCR
        kCR
        
        % sideband wavelength
        kSB
        
        % input mode, defined at addLaser
        % 'LG' or 'HG'
        fieldMode
        % p for LGpm, m for HGmn
        ind1
        % m for LGpm, n for HGmn
        ind2
    end
        
    methods
        
        function obj = fieldManager( ifo )
            obj.ifo = ifo;
            
            obj.fieldList = FieldObj.empty(0);
            
            % CR frequency
            obj.kSB = zeros(1,1);
            
            obj.lamCR = nan;
            obj.kCR = nan;
        end
               
        function xList = getXlist( obj, fieldID )
            xList = obj.fieldList(fieldID).getXlist;
        end
        
        function Wfft = getWfft( obj, fldID )
            Wfft = obj.fieldList(fldID).getWfft;
        end
        
        function [dE, E0, xl] = getPointScatterd( obj, sinkID, u0, v0, du, dv, dh, wOut )
            % [dE, E0, xl] = getPointScatterd( sinkFieldID, srcFieldID, u0, v0, du, dv, dh, wOut )
            % return the 2D field map dE and coordinates xl
            % assuming that srcMirror has a bump at (u0,v0) with widthj of (du, dv) and height dh
            % E.g., sinkFieldID = 'ITMX-HR-i', then defect is placed on ETMX-HR and
            % calculate field on ITM. E0 is the field without defect, and the total field is dE + E0
            
            % convert name to ID
            sinkID = obj.findNamedField( sinkID );
            
            % if not specified, use default
            if ~exist( 'wOut', 'var' )
                wOut = 0;
            end
            
            % field without defect
            [E0, xl] = obj.getField( sinkID, wOut );
            
            % find the source and props
            [srcPort, propID] = obj.ifo.props.getPropFrom( sinkID );
            % prop length
            L0 = obj.ifo.props.getPropLeng(propID);
            % field coming out as the source or reflected field
            srcField = obj.conv2std( srcPort );
            xlSrc = obj.getXlist(srcPort);
            % field at (u0,v0)
            src00 = srcField( find( xlSrc >= v0, 1 ), find( xlSrc >= u0, 1 ) );
            
            % coordinate at sink
            [xg, yg] = meshgrid( xl, xl );
            
            % major part
            dE = du*dv*dh * src00 * (-4*pi/(obj.lamCR^2*L0)) * exp(-1i*obj.kCR*( (xg-u0).^2 + (yg-v0).^2 )/(2*L0));
            
            % minor correction
            % calculate (1/w^2 - i k/2R) from E(w/2,w/2)/E(0,0) = exp(-w^2/2)(1/w^2-i k/2R))
            % Cwr = 1/w^2 - i k /2R = -2/w^2 log( E / E(0,0) )
            % beam size on the source
            wSrc = obj.fieldList(srcPort).w;
            % field at (wSrc/2, wSrc/2) / field(0, 0), chosen for convenience
            srcW = srcField( find( xlSrc >= wSrc/2, 1 ), find( xlSrc >= wSrc/2, 1 ) ) / srcField(end/2,end/2);
            % 1/w^2 - i k / (2R)
            Cwr = -2/wSrc^2*log( srcW );
            
            signhx = @(x) (exp(x)-exp(-x))./(2*x);
            
            % field disturbance by the bump
            correction = signhx( du*(u0*Cwr - 1i*obj.kCR*(xg-u0)/(2*L0)) ) ...
                .* signhx( dv*(v0*Cwr - 1i*obj.kCR*(yg-v0)/(2*L0)) );
            
            dE = dE .* correction;
        end
        
        function setInputType( obj, fieldMode, ind1, ind2 )
            obj.fieldMode = fieldMode;
            obj.ind1 = ind1;
            obj.ind2 = ind2;
        end
        
        function [fieldMode, ind1, ind2] = getInputType( obj )
            fieldMode = obj.fieldMode;
            ind1 = obj.ind1;
            ind2 = obj.ind2;
        end
        
        function gmul = getGouyMul( obj )
            if strcmp(obj.fieldMode, 'LG')
                gmul = 2*obj.ind1 + obj.ind2 + 1;
            else
                gmul = obj.ind1 + obj.ind2 + 1;
            end
        end
        
        function [E, xList] = fieldAtFID( obj, FID, varargin )
            % function [E, xList] = fieldAtFID( fieldID, 'LG or HG', p/m, m/n )
            % This returns 1W field of the type (LG or HG).
            % If called without any arguments, the field type specified at addLaser is used
            
            FID = obj.findNamedField( FID, -1 );
            ip = inputParser;
            addOptional(ip, 'fieldMode', obj.fieldMode, @(x) ischar(x) && (strcmp(x,'HG') || strcmp(x,'LG'))); % field type
            addOptional(ip, 'ind1', obj.ind1, @(x) isnumeric(x) && x>=0 && floor(x) == x ); % first index of the field
            addOptional(ip, 'ind2', obj.ind2, @(x) isnumeric(x) && floor(x) == x ); % second index of the field
            ip.parse(varargin{:})
            
            if strcmp(ip.Results.fieldMode, 'LG')
                [E, xList] = obj.fieldList(FID).LGFieldatFID( ip.Results.ind1, ip.Results.ind2 );
            else
                [E, xList] = obj.fieldList(FID).HGField2DatFID( ip.Results.ind1, ip.Results.ind2 );
            end
        end
                
        function markFieldStatus( obj, fieldID, guessID )
            obj.fieldList(fieldID).status = guessID;
        end
        
        function update = calcIfNecessary( obj, fieldID )
            % if current status == 0, i.e., no field yet ...
            update = obj.fieldList(fieldID).status == 0;
            if update
                if ~obj.ifo.cavities.loopFieldReady || isnan( obj.fieldList(fieldID).invQ )
                    return;
                    %                     error('Lock or calc before calcuating this field');
                end
                obj.fieldList(fieldID).updateField( fieldID, obj.ifo.props, obj.ifo.optics );
            end
        end
        
        function power = calcPower( obj, fieldID )
            power = obj.fieldList(fieldID).calcPower;
        end
        
        function loss = lossOut( obj, fieldID )
        % field power out of a circle optic
        %    loss = 1 - power in aperture / total power in
            fieldID = obj.findNamedField( fieldID, -1 );
            fromPortID = obj.props.getPropFrom( fieldID );
            opticID = obj.ifo.optics.findOptic( fieldID );
            loss = 1 - obj.fieldList(fieldID).calcSumIn(obj.ifo.optics.optList(opticID).Aperture/2) / obj.calcPower(fromPortID);
        end
        
        function [outField, xl, yl] = getField( obj, fieldID, Wout, Lprop )
            % apply propagation by requesting the output window size
            % if Lprop > 0, change the propagation distance

            if ~exist( 'Lprop', 'var' )
                Lprop = 0;
            end
            
            if (Lprop == 0) && (Wout <= obj.getWfft(fieldID)/2) && (Wout > 0)
                % if Wout is small, no need to expand
                outField = obj.conv2std( fieldID );
                xl = obj.getXlist(fieldID);
                yl = xl;
            else
                % if needed, do propagation with large enough target
                outOrIn = 0; % > 0 for out and < 0 for in
                for propID = 1:length(obj.ifo.props.propList)
                    if obj.ifo.props.propList(propID).outPortID == fieldID
                        outOrIn = fieldID;
                        break;
                    elseif obj.ifo.props.propList(propID).inPortID == fieldID
                        outOrIn = -fieldID;
                        break;
                    end
                end

                if outOrIn > 0
                    % end of a propagation, so try to propagate
                    [outField, xl, yl] = ...
                        obj.ifo.props.propList(propID).propOnDemand( Wout, Lprop );               
                    elseif outOrIn < 0
                    % this is a source of a propagation, so just fill 0
                    xl = obj.getXlist(fieldID);
                    Nold = length(xl);
                    dx = xl(2) - xl(1);
                    Nnew = 2*ceil(Wout / dx / 2 );
                    dN2 = (Nnew - Nold)/2;
                    outField = zeros(Nnew);
                    outField( dN2+1:end-dN2, dN2+1:end-dN2 ) = obj.conv2std( fieldID );
                    xl = ((1:Nnew)-Nnew/2-0.5) * dx;
                    yl = xl;
                else
                    error('Field ID %d is not a field', fieldID );
                end
            end
        end
        
        function clearFFTFields( obj, fldIDs )
            if ~exist( 'fldID', 'var' )
                fldIDs = 1:length( obj.fieldList);
            end
            for id = fldIDs
                obj.fieldList(id).EFFT = [];
            end
        end
        
        %   HG coefficient at fldID port, for a standard field Fld
        function Cmn = HGCoef( obj, fldID, m, n, Fld )
            if exist( 'Fld', 'var' )
                Cmn = obj.fieldList(fldID).HGCoef( m, n, Fld );
            else
                Cmn = obj.fieldList(fldID).HGCoef( m, n );
            end
        end
        
        %   LG coefficient at fldID port, for a standard field Fld
        function Cmn = LGCoef( obj, fldID, p, m, Fld )
            if exist( 'Fld', 'var' )
                Cmn = obj.fieldList(fldID).LGCoef( p, m, Fld );
            else
                Cmn = obj.fieldList(fldID).LGCoef( p, m );
            end
        end
        
        % sum of HG(m,n) powers with m+n < mnmax
        function pwrMN = modeSum( obj, fldID, mnmax )
            pwrMN = zeros(1, length(mnmax));
            for imn = 1 : length(mnmax)
                % sum goes from large modes to small modes to minimize the truncation error
                for mn = mnmax(imn) : -1 : 0
                    for m = 0 : mn
                        pwrMN(imn) = pwrMN(imn) + abs(obj.HGCoef( fldID, m, mn-m )).^2;
                    end
                end
            end
        end
        
        function setDefField( obj, fldID, Power, invQ )
            % set the default mode field (specified at addLaser) at port fldID in the internal format
            if ~exist( 'Power', 'var' )
                Power = 1;
            end
            if ~exist( 'invQ', 'var' )
                invQ = obj.fieldList(fldID).invQ;
            end
            % this is a field with the internal format
            %   Einternal = Estandard * exp( -i phiCorr ) * Wfft
            xList = obj.getXlist( fldID );
            [invRoC, w] = MMTK.qinv2RinvW( invQ, obj.kCR );
            weight = sqrt(Power) * obj.getWfft( fldID );
            if strcmp(obj.fieldMode, 'LG')
                obj.fieldList( fldID ).EFFT = ...
                    LGTK.LGFieldRW( obj.ind1, obj.ind2, xList, xList, ...
                    invRoC - obj.fieldList(fldID).phiCorr, w ) * weight;
            else
                obj.fieldList( fldID ).EFFT = ...
                    HGTK.HGField2DRW( obj.ind1, obj.ind2, xList, xList, ...
                    invRoC - obj.fieldList(fldID).phiCorr, w, obj.kCR ) * weight;
            end
        end
        
        function setHGField( obj, fldID, m, n, Power, invQ )
            % set the field at port fldID in the intenal format
            if ~exist( 'Power', 'var' )
                Power = 1;
            end
            if ~exist( 'invQ', 'var' )
                invQ = obj.fieldList(fldID).invQ;
            end
            % this is a field with the internal format
            %   Einternal = Estandard * exp( -i phiCorr ) * Wfft
            xList = obj.getXlist( fldID );
            [invRoC, w] = MMTK.qinv2RinvW( invQ, obj.kCR );
            obj.fieldList( fldID ).EFFT = ...
                sqrt(Power) * HGTK.HGField2DRW( m, n, xList, xList, ...
                invRoC - obj.fieldList(fldID).phiCorr, w, obj.kCR ) * obj.getWfft( fldID );
        end
        
        function invRoC = getInvRoC( obj, fldID )
            invRoC = obj.fieldList(fldID).invRoC;
        end
        
        function Estd = conv2std( obj, fieldID, Eint )
        %   Convert internal field to standard format
            if ~exist( 'Eint', 'var' )
                if isempty( obj.fieldList(fieldID).EFFT )
                    fprintf(1,'Internal field is not defined : %s\n', obj.fieldList(fieldID).name);
                    Estd = obj.ifo.ffts.zeroMat;
                    return
                end
                Eint = obj.fieldList(fieldID).EFFT;
            end
            Estd = obj.fieldList(fieldID).conv2std( Eint );
        end
        
        % create a new field.
        % If coyFromFid is specified,
        %     all data are copifed from the src field and the status is set to 0
        function id = newField( obj, name, copyFromFid )
            % make sure this does not exist
            id = find( strcmp( name, { obj.fieldList(:).name } ), 1 );
            if isempty( id )
                id = length( obj.fieldList ) + 1;
            end
            if exist( 'copyFromFid', 'var' )
                obj.fieldList(id) = copy( obj.fieldList(copyFromFid) );
                obj.fieldList(id).name = name;
            else
                obj.fieldList(id) = FieldObj( name, obj.ifo );
            end
            obj.markFieldStatus(id, 0);
        end
        
        function name = findName( obj, id )
            if (id > 0) && (id <= length(obj.fieldList))
                name = obj.fieldList(id).name;
            else
                name = 'noField';
            end
        end
        
        function id = findNamedField ( obj, name, check )
            % function id = findNamedField ( name, check )
            %
            % Return field id for the field named name
            % If the name is not a valid field name,
            %   if check ~= 0, return id = 0 without interuption
            %   if check == 0, execution exists by error call
            %
            % If name is valid or name is ID number,
            %   if check >= 0, calcIfNecessary is called and, if field is not ready, i.e.,
            %      status is 0, the field is calculated
            %   if check < 0, id is returned without calcIfNecessary
            
            if ~exist( 'check', 'var' )
                check = false;
            end
            if ischar( name )
                id = find( strcmp( name, { obj.fieldList(:).name } ), 1 );
                if isempty( id )
                    id = find( strcmp( name, { obj.ifo.optics.optList(:).name } ), 1 );
                    if ~isempty( id )
                        if obj.ifo.optics.optList(id).isLaserObj
                            id = obj.ifo.optics.findPortID( sprintf('%s-1-o', name) );
                        elseif obj.ifo.optics.optList(id).isDetectorObj
                            id = obj.ifo.optics.findPortID( sprintf('%s-1-i', name) );
                        else
                            id = [];
                        end
                    end
                    if isempty( id )
                        id = 0;
                        if ~check
                            error('Field name ''%s'' does not exist', name );
                        end
                        return;
                    end
                end
            else
                id = name;
            end
            
            % if check < 0, don't do anything
            if check >= 0
                % if the field is not ready, i.e., status == 0, calculate using available fields
                obj.calcIfNecessary( id );
            end
        end
        
        function setFieldName( obj, id, name )
            if (id>0) && (length(obj.fieldList)>=id)
                obj.fieldList(id).name = name;
            end
        end
        
        function phiCorr = getPhiCorr( obj, fieldID )
            phiCorr = obj.fieldList( fieldID ).phiCorr;
        end
        
        % print fields
        % if all ia false, only fields with power are printed
        function printAll( obj, fid )
            if ~exist( 'fid', 'var' )
                fid = 1;
            end
            fileOpened = ischar( fid );
            if fileOpened
                fid = fopen(fid, 'w');
            end
            
            fprintf( fid, '\n=== Information about fields ===\n\n');
            for ii = 1:length(obj.fieldList)
                info =  obj.fieldList(ii).printInfo( true );
                if ~isempty( info )
                    fprintf( fid, '%d) %s\n\n', ii, info );
                end
            end
            
            if fileOpened
                fclose(fid);
            end
        end
        
    end
    
    methods (Static)
    end
end
