% function dat = fit1dData(filename, xl, yl, rOut, indMax)
%
% load 2 column (x, z) and return data at (xl, yl).
% In the outer region with r larger than the region given in the input is filled by extrapolation
% by using extendOuterside( indMax );
%
% data needs to be given either in one side (0, 1, 2, 3...) or in two sides symmetrically, (-3,-2,-1,0,1,2,3)
%
% 2 column
%    x1, z1
%    x2, z2
%    ...
%    if x1 >= 0, xi is used as ri and zi is the value at ri
%    if x1 = -x1, data on full area is generated by interpolating around phi,
%    with two boundary condition, z(ri, phi=0) = z at xi=ri, z(ri,phi=pi) = z at xi=-ri
%
% Hiro Yamamoto
%
% April 29, 2015  'if x(1) == 0, assume x = r' is changed to if x(1) >= 0, assume x = r'
%
% December 16, 2014  Original code
%
function datOut = fit1dData(filename, xl, yl, rOut, indMax)

% open the file
[fid, message]=fopen(filename,'r');
if fid == -1
    error('File(''%s'') open failed because %s', filename, message);
end

while true
    fpos = ftell( fid );
    aLine = fgets(fid);
    [~,numColumns] = sscanf( aLine, '%g');
    if numColumns == 2
        fseek( fid, fpos, 'bof' );
        datIn = fscanf( fid, '%g', [numColumns, inf] );
        break;
    elseif numColumns > 2
        error('Too many data (%d) in one line', numColumns );
    end
end

if size(xl,2) == 1
    xl = xl';
end
if size(yl,1) == 1
    yl = yl';
end
[xl, yl] = meshgrid( xl(1,:), yl(:,1) );
datOut = nan( size(xl) );

x = datIn(1,:);
z = datIn(2,:);
r = sqrt( xl.^2 + yl.^2 );
rmax = max( abs(x) );
indR = find( r < rmax );
if x(1) >= 0
    % x = r
    datOut(indR) = interp1( x, z, r(indR) );
elseif isempty( find(x(1:floor(end/2)) ~= -x(end:-1:ceil(length(x)/2)+1),1) )
    % [-3,-2,-1,0,1,2,3] like x value
    theta = abs(atan2( yl, xl ));
    rp = zeros( size(xl) );
    rm = zeros( size(xl) );
    rp(indR) = interp1( x(ceil(end/2):end), z(ceil(end/2):end), r(indR));
    rm(indR) = interp1( x(ceil(end/2):end), z(ceil(end/2):-1:1), r(indR));
    datOut(indR) = (rm(indR)-rp(indR))/pi.*theta(indR) + rp(indR);
end
fclose(fid);

if ~exist('indMax', 'var')
    indMax = 0;
end
datOut = extendOuterside( datOut, xl, yl, rmax, rOut, indMax );
