%% Simple FP cavity simulation, using LLO Y arm optics
%
% minimal template
%
% copy the following in a file IFOname.m, whete IFOname can be anything
% and change IFOname to whatever you used for the file name, like aLIGO or
% XarmFP
%
% classdef IFOname < FFTIFO
%     methods
%         function obj = IFOname( varargin )
%             obj@FFTIFO( varargin{:} );
%         end
% 
%         function defineIFO( obj, varargin )
%             eval( UtilTK.setVars( who, true, varargin{:} ) );
%         end
%     end
% end

%
classdef FPIFO < FFTIFO
    
    methods
        %% this definition allows to pass parameters at runtime, like
        %  fp = FPIFO('tiltE', 1e-8);
        function obj = FPIFO( varargin )
            obj@FFTIFO( varargin{:} );
        end

        %% this defines the IFO configuration
        %  1) Use addLaser, addMirror and addDetector to define the optical
        %  configuration.
        %  2) Use setHRfiles to define the details of the HR surface
        %  2-1)  Use the third argument of setHRfile to add aberration
        %        using analytic form, like
        %     'map + y*tilt' to add a tilt or
        %     'map + (x.^2+y.^2)/(2+RRH) to add a power change

        function defineIFO( obj, varargin )
            armFP = 'LY';
            eval( UtilTK.setVars( 'FP', true, varargin{:} ) );

            aLIGOsetup;
            
            % input beam, HG(ind1, ind2) or LG(ind1, ind2)
            fieldMode = 'HG';
            ind1 = 0;
            ind2 = 0;
            
            % tilt of mirrors
            tiltI = 0;
            tiltE = 0;

            % cavity length
            Lcav = 3994.5;

            % cavity 
            Aetm = 0.326;
            Aitm = Aetm;
            lossMul = 1;
            
            extraArmLoss = 40e-6;
            removeCurvature = 0;
            resolution = 2e-3;

            % modify maps
            mapMul = 1;
            aidanMul = 0;
            aidanMap = '/Users/hiro/Dropbox/RichardDay/FOGPrime13/SingleMode/Aidan/displacementSurfacePI_120mm_1W.txt';
            PcoatI = 0;
            PcoatE = 0;

            % parameters can be changed by runtime arguments, like fp = FPIFO('tiltE', 1e-8)
            eval( UtilTK.setVars( who, true, varargin{:} ) );

            %% add optics
            % laser source
            obj.addLaser( 'src', 'fieldMode', fieldMode, 'ind1', ind1, 'ind2', ind2 );
            
            % ITM
            obj.addMirror( 'ITM', 'invRoC', 1/ITMRoC, 'Aperture', Aitm, 'Thick', 0.2, ...
                'T', TITM, 'Lhr', Loss_I_HR*lossMul + extraArmLoss/2, 'Lar', Loss_I_AR*lossMul, 'removeTilt', false, 'removeCurvature', removeCurvature );
            % ETM
            ETMID = obj.addMirror( 'ETM', 'invRoC', 1/ETMRoC, 'Aperture', Aetm, 'Thick', 0.2, ...
                'T', TETM, 'Lhr', Loss_E_HR*lossMul + extraArmLoss/2, 'resolution', resolution,  'removeTilt', false, 'removeCurvature', removeCurvature  );
            
            % detector behind ETM
            obj.addDetector( 'trans' );
            
            %% define connections among optics
            obj.addProp( '', 'src', 'ITM-AR' );
            obj.addProp( 'FPprop', 'ITM-HR', 'ETM-HR', Lcav);
            obj.addProp( '', 'ETM-AR', 'trans' );
            
            %% find beam size on ITM and ETM
            if PcoatI > 0 || PcoatE > 0
                [wI, wE] = obj.findBeamSize;
            end
            
            %% use the measured maps
            % ITM HR surface
            if ~isempty( ITMmap )
                obj.setHRfiles( 'ITM', ITMmap, 'mapMul * map+y*tiltI', 0.16, ITMRoC, 0, 'tiltI', tiltI, 'mapMul', mapMul  );
            end
            % ETM HR surface
            if ~isempty( ETMmap )
                obj.setHRfiles( 'ETM', ETMmap, 'mapMul * map+y*tiltE', 0.16, ETMRoC, 0, 'tiltE', tiltE, 'mapMul', mapMul );
            end
            
            if ~isempty( aidanMap )
                obj.setHRfiles( 'ITM', '', strcat('-aidanMul * fit1dData(''',aidanMap, ''', xl, yl, 0.17)'), 0.16, 0, 0, 'aidanMul', aidanMul );
                obj.setHRfiles( 'ETM', '', strcat('-aidanMul * fit1dData(''',aidanMap, ''', xl, yl, 0.17)'), 0.16, 0, 0, 'aidanMul', aidanMul );
            end
            if PcoatI > 0
                obj.setHRfiles( 'ITM', '', 'calcThermal(0, r, 0, PcoatI, wI, 0.2, 0.17 )', 0.16, 0, 0, 'PcoatI', PcoatI, 'wI', wI );
            end

            if PcoatE > 0
                obj.setHRfiles( 'ETM', '', 'calcThermal(0, r, 0, PcoatE, wE, 0.2, 0.17 )', 0.16, 0, 0, 'PcoatE', PcoatE, 'wE', wE );
            end

        end

        function [wI, wE] = findBeamSize( obj )
            obj.analyzeCavity;
            obj.setupBaseMode;
            wI = obj.fields.fieldList( obj.fields.findNamedField( 'ITM-HR-o', -1 ) ).w;
            wE = obj.fields.fieldList( obj.fields.findNamedField( 'ETM-HR-o', -1 ) ).w;
        end

    end
end
