%% FP cavity to study performance using various map combination
%
% example usage
%{
% build FP cavity, change values IF you need to, want to
fp = RanaFP( 'ITMmap', './data/xITM_11_R1_Figure.dat', ...
             'ETMmap', './data/xETM_11_R1_Figure.dat', ...
             'ITMRoC', 1934, 'ITMAper', 0.326, ...
             'ETMRoC', 2245, 'ETMAper', 0.326, 'rotateE', 1, ...
             'resolution', 0.002, 'lamCR', 1.064e-6 );
% lock the cavity
fp.lock;
% get the round trip loss
fp.roundtripLoss
% print LG fraction in the cavity, LG00, LG10, ..., LG50
fp.LGFrac('ETM-HR-o', 0:5,0 )
% see field, on ITM, HR side, incoming field
[E, xy] = fp.getField('ITM-HR-i');
pcolor( xy, xy, log(abs(E).^2) ); shading flat; axis equal; colorbar;
% power loss on ETM
1 - fp.power('ETM-HR-o')/fp.power('ETM-HR-i')
% see information
help fp.addMirror
% save cavity information in a file
fp.printAll('RanaFP.log');
%}
%
classdef RanaFP < FFTIFO
    
    methods
        function obj = RanaFP( varargin )
            obj@FFTIFO( varargin{:} );
        end

        function defineIFO( obj, varargin )
            % default values of cavity parameters, which can be changed

            Lcav = 3994.5;
            ITMRoC = 1934; ITMAper = 0.326; T_ITM = 0.014;
            ETMRoC = 2234; ETMAper = 0.326; T_ETM = 5e-6;
            ITMmap = './data/xITM_11_R1_Figure.dat';
            ETMmap = './data/xETM_11_R1_Figure.dat';
            resolution = 0.002;            
            lamCR = 1.064e-6;
            rotateI = 0; % rotate ITM map by pi/4 x rotateI
            rotateE = 0;
            
            % parameters can be changed by runtime arguments, like fp = FPIFO('tiltE', 1e-8)
            eval( UtilTK.setVars( who, true, varargin{:} ) );

            %% add optics
            % laser source
            obj.addLaser( 'src', 'lamCR', lamCR );
            
            % ITM
            obj.addMirror( 'ITM', 'invRoC', 1/ITMRoC, 'Aperture', ITMAper, 'Thick', 0.2, ...
                'T', T_ITM, 'Lhr', 0, 'Lar', 0 );
            % ETM
            obj.addMirror( 'ETM', 'invRoC', 1/ETMRoC, 'Aperture', ETMAper, 'Thick', 0.2, ...
                'T', T_ETM, 'Lhr', 0, 'resolution', resolution );
            
            % detector behind ETM
            obj.addDetector( 'trans' );
            
            %% define connections among optics
            obj.addProp( '', 'src', 'ITM-AR' );
            obj.addProp( 'FPprop', 'ITM-HR', 'ETM-HR', Lcav);
            obj.addProp( '', 'ETM-AR', 'trans' );
            
          
            %% use the measured maps
            % ITM HR surface
            obj.setHRfiles( 'ITM', ITMmap, '', 0.16, ITMRoC, rotateI );

            % ETM HR surface
            obj.setHRfiles( 'ETM', ETMmap, '', 0.16, ETMRoC, rotateE );
        end
    end
end
